﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Infragistics.Win.UltraWinTree;

namespace WindowMaxMinDockMgr
{
    public partial class DockStructureObserver : Form
    {
        #region Member variables
        private readonly DockerPanel m_DockerPanel;
        private readonly List<LeafNodeMenuItemMetaData> m_LeafNodeContextMenuMetaData;
        #endregion Member variables

        #region Helper classes
        internal class LeafNodeMenuItemMetaData
        {
            private readonly Func<object, bool> _test;
            private readonly string _text;
            private Func<object, string> _getText;
            private readonly EventHandler _handler;
            internal LeafNodeMenuItemMetaData(Func<object, bool> test,
                string text, EventHandler handler)
            {
                _test = test;
                _text = text;
                _getText = null;
                _handler = handler;
            }
            internal LeafNodeMenuItemMetaData(Func<object, bool> test,
                Func<object, string> getText, EventHandler handler)
            {
                _test = test;
                _text = null;
                _getText = getText;
                _handler = handler;
            }
            internal Func<object, bool> Test { get { return _test; } }
            internal string Text { get { return _text; } }
            internal Func<object, string> GetText { get { return _getText; } }
            internal bool HasText { get { return _text != null; } }
            internal EventHandler Handler { get { return _handler; } }
        }
        #endregion Helper classes

        #region Constructor
        public DockStructureObserver(DockerPanel dockerPanel, bool alwaysOnTop)
        {
            InitializeComponent();

            m_DockerPanel = dockerPanel;
            dockerPanel.DockChange += new EventHandler(dockerPanel_DockChange);
            dockerPanel.PaneActivate += new DockerPanel.PaneEventHandler(DockerPanel_PaneActivate);

            m_LeafNodeContextMenuMetaData = GetLeafNodeContextMenuMetaData();

            ActivatingNodeFromPane = false;
            PreserveSelectedNodes = false;

            Tree.HideSelection = false;
            Tree.SelectionBehavior = SelectionBehavior.ExtendedAcrossCollections;

            Tree.AfterActivate += new AfterNodeChangedEventHandler(ultraTree1_AfterActivate);
            Tree.MouseClick += new MouseEventHandler(ultraTree1_MouseClick);
            Tree.MouseDoubleClick += new MouseEventHandler(ultraTree1_MouseDoubleClick);
            Tree.MouseUp += new MouseEventHandler(ultraTree1_MouseUp);

            this.TopMost = alwaysOnTop;
        }
        #endregion Constructor

        #region Properties
        private DockerPanel DockerPanel { get { return m_DockerPanel; } }
        private UltraTree Tree { get { return this.ultraTree1; } }


        private bool m_PreserveSelectedNodes;
        private bool PreserveSelectedNodes
        {
            get { return m_PreserveSelectedNodes; }
            set 
            {
                Debug.WriteLine("PreserveSelectedNodes=" + value.ToString());
                m_PreserveSelectedNodes = value;
            }
        }

        private bool m_ActivatingNodeFromPane;
        private bool ActivatingNodeFromPane
        {
            get { return m_ActivatingNodeFromPane; }
            set
            {
                Debug.WriteLine("ActivatingNodeFromPane=" + value.ToString());
                m_ActivatingNodeFromPane = value;
            }
        }

        private List<LeafNodeMenuItemMetaData> LeafNodeContextMenuMetaData { get { return m_LeafNodeContextMenuMetaData; } }
        #endregion Properties

        #region Methods
        private List<LeafNodeMenuItemMetaData> GetLeafNodeContextMenuMetaData()
        {
            List<LeafNodeMenuItemMetaData> metaData = new List<LeafNodeMenuItemMetaData>()
            {
                new LeafNodeMenuItemMetaData((p)=>{return DockerPanel.IsPaneMinimized(p);}, "Restore", RestoreNode_Click),
                new LeafNodeMenuItemMetaData((p)=>{return DockerPanel.IsPaneMinimized(p);}, "Maximize", MaximizeNode_Click),

                new LeafNodeMenuItemMetaData((p)=>{return DockerPanel.IsPaneMaximized(p);}, "Restore", RestoreNode_Click),
                new LeafNodeMenuItemMetaData((p)=>{return DockerPanel.IsPaneMaximized(p);}, "Minimize", MinimizeNode_Click),

                new LeafNodeMenuItemMetaData((p)=>{return DockerPanel.IsPaneNormal(p);}, "Minimize", MinimizeNode_Click),
                new LeafNodeMenuItemMetaData((p)=>{return DockerPanel.IsPaneNormal(p);}, "Maximize", MaximizeNode_Click),
                new LeafNodeMenuItemMetaData((p)=>{return DockerPanel.IsPaneNormal(p);}, 
                    (p)=> { return (DockerPanel.IsPaneDocked(p) ? "Undock" : "Dock"); }, 
                    ToggleDockStateNode_Click),
                new LeafNodeMenuItemMetaData((p)=>{return DockerPanel.IsPaneNormal(p);}, 
                    "Tile selected documents horizontally", TileSelectedNodesHorizontally_Click),
                new LeafNodeMenuItemMetaData((p)=>{return DockerPanel.IsPaneNormal(p);}, 
                    "Tile selected documents vertically", TileSelectedNodesVertically_Click),

                // These items will be added to every context menu because Test() always returns true.
                new LeafNodeMenuItemMetaData((p)=>{return true;}, "Close", CloseNode_Click)
            };
            return metaData;
        }

        private bool IsLeafNode(UltraTreeNode node)
        {
            return !node.HasNodes;
        }

        private void ActivateNodeForPane(UltraTree tree, object pane)
        {
            Utilities.NestedTrace(() =>
                {
                if (pane == null)
                {
                    Debug.WriteLine("No pane so no node to activate");
                    return;
                }

                UltraTreeNode node = FindNodeByPane(tree.Nodes, pane);
                if (node != null)
                {
                    tree.ActiveNode = node;
                    if (!PreserveSelectedNodes) tree.SelectedNodes.Clear();
                    tree.ActiveNode.Selected = true;
                }
                else
                {
                    Debug.WriteLine("No node for pane");
                }
            });
        }

        private UltraTreeNode FindNodeByPane(TreeNodesCollection nodes, object pane)
        {
            foreach (UltraTreeNode node in nodes)
            {
                UltraTreeNode foundNode = FindNodeByPane(node, pane);
                if (foundNode != null) return foundNode;
            }
            return null;
        }

        private UltraTreeNode FindNodeByPane(UltraTreeNode node, object pane)
        {
            if (PaneFromNode(node) == pane) return node;
            return FindNodeByPane(node.Nodes, pane);
        }

        private object PaneFromNode(UltraTreeNode node) { return node.Tag; }
        private void SetPaneForNode(UltraTreeNode node, object pane) { node.Tag = pane; }

        private List<object> PanesFromSelectedNodes(UltraTree tree, Func<UltraTreeNode, bool> filter)
        {
            List<object> panes = new List<object>();
            foreach (UltraTreeNode node in tree.SelectedNodes)
                if (filter(node)) panes.Add(PaneFromNode(node));
            return panes;
        }

        private List<object> PanesFromSelectedLeafNodes(UltraTree tree)
        {
            return PanesFromSelectedNodes(tree, (node) => { return IsLeafNode(node); });
        }

        private void WithSelectedNodesPreserved(Action action)
        {
            try
            {
                PreserveSelectedNodes = true;
                action();
            }
            finally
            {
                PreserveSelectedNodes = false;
            }
        }

        private void SelectTab(UltraTreeNode node)
        {
            Utilities.NestedTrace(() =>
            {
                DockerPanel.SelectTab(PaneFromNode(node));
            });
        }

        private void SafeSelectAndMaybeActivate(UltraTreeNode node, bool giveFocus)
        {
            Utilities.NestedTrace(() =>
            {
                if (!m_ActivatingNodeFromPane && node != null)
                {
                    object pane = PaneFromNode(node);

                    if (DockerPanel.IsPaneMinimized(pane))
                    {
                        DockerPanel.RestorePane(pane);
                    }
                    else if (DockerPanel.IsPaneNormal(pane) && !DockerPanel.IsPaneVisible(pane))
                    {
                        DockerPanel.RestorePane(pane);
                    }
                    else if (!DockerPanel.IsPaneVisible(pane))
                    {
                        Debug.Assert(DockerPanel.IsPaneMaximized(pane));
                        Debug.Assert(DockerPanel.PaneHasProxyForm(pane));
                        DockerPanel.GetProxyForm(pane).Show();
                    }
                    SelectTab(node);
                    if (giveFocus) DockerPanel.GiveFocus(pane);
                }
            });
        }

        private void SafeMaybeActivate(bool giveFocus)
        {
            SafeSelectAndMaybeActivate(Tree.ActiveNode, giveFocus);
        }

        public void Initialize()
        {
            Utilities.NestedTrace(() =>
            {
                Tree.Nodes.Clear();
                DockerPanel.WithEachTopLevelPane((p) => { InitializeSubTree(Tree.Nodes, p); });
            });
        }

        private void Reinitialize()
        {
            Utilities.NestedTrace(() =>
            {
                List<object> panes = PreserveSelectedNodes
                    ? PanesFromSelectedLeafNodes(Tree)
                    : new List<object> { DockerPanel.GetActivePane() };

                Initialize();
                Tree.ExpandAll();
                ActivateNodeForPane(Tree, DockerPanel.GetActivePane());
                SelectPanes(Tree, panes);
            });
        }

        private void SelectPanes(UltraTree tree, List<object> panes)
        {
            tree.SelectedNodes.Clear();
            foreach (object pane in panes) SelectPane(tree, pane);
        }

        private void SelectPane(UltraTree tree, object pane)
        {
            UltraTreeNode node = FindNodeByPane(tree.Nodes, pane);
            if (node != null) node.Selected = true;
        }

        private void InitializeSubTree(TreeNodesCollection nodes, object pane)
        {
            if (DockerPanel.IsEmptyGroup(pane))
            {
                Debug.WriteLine("InitializeSubTree: '" + 
                    DockerPanel.GetPaneText(pane) + "' is empty");
                return;
            }

            UltraTreeNode node = nodes.Add();
            SetPaneForNode(node, pane);
            node.Text = DockerPanel.GetPaneText(pane);

            DockerPanel.WithEachChildPane(pane, (p) => { InitializeSubTree(node.Nodes, p); });
        }

        private ContextMenuStrip GetContextMenuForTree()
        {
            ContextMenuStrip menu = new ContextMenuStrip();

            ToolStripItem itemA = menu.Items.Add("Tree Action A");
            itemA.Click += new EventHandler(treeA_Click);

            ToolStripItem itemB = menu.Items.Add("Tree Action B");
            itemB.Click += new EventHandler(treeB_Click);

            return menu;
        }

        private object PaneFromMenuItem(ToolStripItem menuItem)
        {
            return menuItem.Tag;
        }

        private void SetPaneForMenuItem(ToolStripItem menuItem, object pane)
        {
            menuItem.Tag = pane;
        }

        private ContextMenuStrip GetContextMenuForNode(UltraTreeNode node)
        {
            Debug.Assert(IsLeafNode(node));

            // Offer menu commands based on state of associated pane.
            object pane = PaneFromNode(node);
            Debug.Assert(pane != null);

            ContextMenuStrip menu = new ContextMenuStrip();
            foreach (LeafNodeMenuItemMetaData imd in m_LeafNodeContextMenuMetaData)
            {
                if (imd.Test(pane))
                {
                    string text = imd.HasText ? imd.Text : imd.GetText(pane);
                    ToolStripItem item = menu.Items.Add(text);
                    SetPaneForMenuItem(item, pane);
                    item.Click += imd.Handler;
                }
            }
            return menu;
        }

        private object GetPaneFromMenuClick(object sender)
        {
            ToolStripMenuItem menuItem = sender as ToolStripMenuItem;
            Debug.Assert(menuItem != null);
            object pane = PaneFromMenuItem(menuItem);
            Debug.Assert(pane != null);
            return pane;
        }

        private void TileSelectedNodes(Func<int, Point, Size, List<Tuple<Point, Size>>> getPositions)
        {
            List<object> panes = PanesFromSelectedLeafNodes(Tree);
            Rectangle wa = Screen.PrimaryScreen.WorkingArea;
            DockerPanel.TilePanes(panes, wa.Location, wa.Size, getPositions);
            Reinitialize();
        }

        #endregion Methods

        #region Event handlers
        void DockerPanel_PaneActivate(object sender, DockerPanel.PaneEventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                if (ActivatingNodeFromPane) return;

                try
                {
                    ActivatingNodeFromPane = true;
                    ActivateNodeForPane(Tree, e.Pane);
                }
                finally
                {
                    ActivatingNodeFromPane = false;
                }
            });
        }

        void ultraTree1_MouseClick(object sender, MouseEventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                // Was it a left-click on a leaf node?
                // If yes, select the associated pane.
                if (e.Button == MouseButtons.Left)
                {
                    UltraTreeNode node = Tree.GetNodeFromPoint(e.Location);
                    if (node != null && IsLeafNode(node))
                    {
                        SafeSelectAndMaybeActivate(node, false);
                    }
                }
            });
        }

        void ultraTree1_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                // Was the double-click on a leaf node?
                // If yes, select the associated pane and give it the focus.
                UltraTreeNode node = Tree.GetNodeFromPoint(e.Location);
                if (node != null && IsLeafNode(node))
                {
                    SafeSelectAndMaybeActivate(node, true);
                }
            });
        }

        void ultraTree1_AfterActivate(object sender, NodeEventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                if (!ActivatingNodeFromPane)
                {
                    // The tree was activated directly by the user.
                    if (Tree.ActiveNode != null)
                    {
                        SelectTab(Tree.ActiveNode);
                    }
                }
                else
                {
                    // The tree was activated indirectly by, for example,
                    // creation of a new pane.
                    Debug.WriteLine("Indirect activation");
                }
            });
        }

        void dockerPanel_DockChange(object sender, EventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                Reinitialize();
            });
        }

        void ultraTree1_MouseUp(object sender, MouseEventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                if (e.Button == MouseButtons.Right)
                {
                    Debug.WriteLine("\tRight-button click");

                    // Determine the context menu to show based on the location of the mouse event.
                    // If over a node, display a menu for that node. Otherwise, display a menu for
                    // the whole tree.

                    ContextMenuStrip menu = null;
                    UltraTreeNode nodeAtPoint = Tree.GetNodeFromPoint(e.Location);
                    if (nodeAtPoint != null)
                    {
                        if (IsLeafNode(nodeAtPoint))
                        {
                            Debug.WriteLine("\tClick was on leaf node");
                            menu = GetContextMenuForNode(nodeAtPoint);
                        }
                        else
                        {
                            Debug.WriteLine("\tClick was on group node");
                        }
                    }
                    else
                    {
                        Debug.WriteLine("\tClick was not on node");
                        menu = GetContextMenuForTree();
                    }

                    if (menu != null)
                    {
                        menu.Show(sender as Control, e.Location);
                    }
                }
            });
        }

        void RestoreNode_Click(object sender, EventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                WithSelectedNodesPreserved(() =>
                {
                    DockerPanel.RestorePane(GetPaneFromMenuClick(sender));
                });
            });
        }

        void ToggleDockStateNode_Click(object sender, EventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                DockerPanel.ToggleDockState(GetPaneFromMenuClick(sender));
            });
        }

        void TileSelectedNodesHorizontally_Click(object sender, EventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                WithSelectedNodesPreserved(() => 
                { 
                    TileSelectedNodes(Utilities.SplitHorizontally);
                });
            });
        }

        void TileSelectedNodesVertically_Click(object sender, EventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                WithSelectedNodesPreserved(() => 
                {
                    TileSelectedNodes(Utilities.SplitVertically); 
                });
            });
        }

        void MaximizeNode_Click(object sender, EventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                DockerPanel.MaximizePane(GetPaneFromMenuClick(sender), true);
            });
        }

        void MinimizeNode_Click(object sender, EventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                DockerPanel.MinimizePane(GetPaneFromMenuClick(sender), true);
            });
        }

        void CloseNode_Click(object sender, EventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                WithSelectedNodesPreserved(() => 
                {
                    DockerPanel.DestroyPane(GetPaneFromMenuClick(sender));
                    Reinitialize();
                });
            });
        }

        void treeA_Click(object sender, EventArgs e)
        {
            MessageBox.Show("Tree Action A");
        }

        void treeB_Click(object sender, EventArgs e)
        {
            MessageBox.Show("Tree Action B");
        }
        #endregion Event handlers
    }
}
