﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Infragistics.Win.UltraWinDock;

namespace WindowMaxMinDockMgr
{
    public static class DockableControlPaneExtensions
    {
        public static PaneState GetPaneStateObject(this DockableControlPane pane)
        {
            PaneState paneState = pane.Tag as PaneState;
            Debug.Assert(paneState != null);
            return paneState;
        }

        public static PaneState.PaneStateValue GetPaneState(this DockableControlPane pane)
        {
            PaneState state = pane.GetPaneStateObject();
            return state.Value;
        }

        public static PaneProxyForm GetProxyForm(this DockableControlPane pane)
        {
            PaneState state = pane.GetPaneStateObject();
            return state.ProxyForm;
        }

        public static void SetPaneState(this DockableControlPane pane, PaneState.PaneStateValue state, DockerPanel dockerPanel)
        {
            PaneState paneState = new PaneState(state, dockerPanel);
            pane.Tag = paneState;
        }

        public static void UpdatePaneState(this DockableControlPane pane, PaneState.PaneStateValue newStateValue)
        {
            PaneState oldState = pane.GetPaneStateObject();
            pane.Tag = new PaneState(oldState, newStateValue);
        }

        public static void UpdatePaneState(this DockableControlPane pane, PaneProxyForm proxyForm)
        {
            PaneState oldState = pane.GetPaneStateObject();
            pane.Tag = new PaneState(oldState, proxyForm);
        }

        public static bool IsPaneMinimized(this DockableControlPane pane)
        {
            PaneState paneState = pane.GetPaneStateObject();
            return paneState.IsMinimized;
        }

        public static bool IsPaneMaximized(this DockableControlPane pane)
        {
            PaneState paneState = pane.GetPaneStateObject();
            return paneState.IsMaximized;
        }

        public static bool IsPaneNormal(this DockableControlPane pane)
        {
            PaneState paneState = pane.GetPaneStateObject();
            return paneState.IsNormal;
        }

        public static bool HasProxyForm(this DockableControlPane pane)
        {
            PaneState paneState = pane.GetPaneStateObject();
            return paneState.HasProxyForm;
        }

        public static void EnsureFormExists(this DockableControlPane pane)
        {
            if (!pane.HasProxyForm())
            {
                PaneState paneState = pane.GetPaneStateObject();
                PaneProxyForm form = new PaneProxyForm(pane, paneState.DockerPanel.HideMinimizedWindows);
                form.PaneResize += new EventHandler(form_PaneResize);
                form.DestroyingForm += new EventHandler(form_DestroyingForm);
                form.Text = pane.Text;
                form.ShowInTaskbar = false;
                pane.UpdatePaneState(form);
            }
        }

        private static void form_PaneResize(object sender, EventArgs e)
        {
            PaneProxyForm form = sender as PaneProxyForm;
            DockableControlPane pane = form.Pane as DockableControlPane;
            pane.UpdatePaneStateFromWindowState(form.WindowState);
        }

        private static void form_DestroyingForm(object sender, EventArgs e)
        {
            PaneProxyForm form = sender as PaneProxyForm;
            form.PaneResize -= new EventHandler(form_PaneResize);
            form.DestroyingForm -= new EventHandler(form_DestroyingForm);
        }

        public static XPanel GetXPanel(this DockableControlPane pane)
        {
            XPanel panel = pane.Control as XPanel;
            Debug.Assert(panel != null);
            return panel;
        }

        public static void ContentToFormFromPane(this DockableControlPane pane, PaneProxyForm form)
        {
            XPanel panel = pane.GetXPanel();
            Control content = panel.GetContent();
            panel.ClearContent();
            form.SetContent(content);
        }

        public static void ContentToPaneFromForm(this DockableControlPane pane, PaneProxyForm form)
        {
            XPanel panel = pane.GetXPanel();
            Control content = form.GetContent();
            form.ClearContent();
            panel.SetContent(content);
        }

        public static void ActivatePane(this DockableControlPane pane)
        {
            Debug.WriteLine("ActivatePane");
            // If control is maximized, don't do anything.
            if (pane.IsPaneMaximized())
            {
                Debug.WriteLine("\tWas maximized");
                Debug.Assert(pane.HasProxyForm());
                PaneProxyForm form = pane.GetProxyForm();
                form.Activate();
            }
            else if (pane.IsPaneMinimized())
            {
                Debug.WriteLine("\tWas minimized");
                pane.RestorePane();
            }
            else
            {
                Debug.Assert(pane.IsPaneNormal());
                Debug.WriteLine("\tWas normal");

                // Control is a docked or floating window. Make it visible.
                pane.Show();
                pane.Activate();
            }
        }

        private static void HandleMinMaxButton(this DockableControlPane pane, PaneState.PaneStateValue newState, 
            bool showForm, Action<DockableControlPane, PaneState.PaneStateValue, PaneProxyForm> action)
        {
            PaneState.PaneStateValue oldState = pane.GetPaneState();
            if ((oldState == PaneState.PaneStateValue.Minimized) &&
                (newState == PaneState.PaneStateValue.Minimized))
            {
                // Probably changing "hide on minimize" and windows are being displayed.
            }
            else
            {
                Debug.Assert(oldState != newState);
            }

            pane.EnsureFormExists();
            PaneProxyForm form = pane.GetProxyForm();
            form.Resizing = true;
            Debug.Assert(form != null);
            form.State = newState;
            pane.UpdatePaneState(newState);
            action(pane, oldState, form);
            form.Resizing = false;

            Debug.WriteLine("HandleMinMaxButton (leaving): " + form.DumpState());
        }

        public static void MinimizePane(this DockableControlPane pane, bool showForm)
        {
            HandleMinMaxButton(pane, PaneState.PaneStateValue.Minimized, showForm,
                (controlPane, oldState, form) =>
                {
                    Debug.WriteLine("MinimizePane: " + form.DumpState());
                    if (oldState == PaneState.PaneStateValue.Normal)
                    {
                        controlPane.Close();
                        controlPane.ContentToFormFromPane(form);
                    }
                    else
                    {
                        controlPane.Close();
                        Debug.Assert(oldState == PaneState.PaneStateValue.Maximized);
                        Debug.Assert(form.GetContent() != null);
                    }

                    if (showForm && !form.HideWhenMinimized) form.Show();

                });
        }

        public static void MaximizePane(this DockableControlPane pane, bool showForm)
        {
            HandleMinMaxButton(pane, PaneState.PaneStateValue.Maximized, showForm,
                (controlPane, oldState, form) =>
                {
                    Debug.WriteLine("MaximizePane: " + form.DumpState());
                    if (oldState == PaneState.PaneStateValue.Normal)
                    {
                        controlPane.Close();
                        controlPane.ContentToFormFromPane(form);
                    }
                    else
                    {
                        Debug.Assert(!controlPane.IsVisible);
                        Debug.Assert(oldState == PaneState.PaneStateValue.Minimized);
                        Debug.Assert(form.GetContent() != null);
                    }

                    if (showForm) form.Show();
                });
        }

        public static void RestorePane(this DockableControlPane pane)
        {
            if (pane.IsPaneNormal())
            {
                // We're making visible a pane that was hidden.
                Debug.Assert(!pane.IsVisible);
                pane.Show();
            }
            else
            {
                Debug.Assert(pane.HasProxyForm());
                PaneProxyForm form = pane.GetProxyForm();
                form.Resizing = true;
                form.Hide();
                pane.ContentToPaneFromForm(form);
                pane.Show();

                pane.UpdatePaneState(PaneState.PaneStateValue.Normal);

                form.Resizing = false;
            }
        }

        // The proxy form was resized.
        public static void UpdatePaneStateFromWindowState(this DockableControlPane pane, FormWindowState windowState)
        {
            if (pane.HasProxyForm() && !pane.GetProxyForm().Resizing)
            {
                if (windowState == FormWindowState.Normal)
                {
                    pane.RestorePane();
                }
                else if (windowState == FormWindowState.Minimized)
                {
                    pane.MinimizePane(false);
                }
                else
                {
                    Debug.Assert(windowState == FormWindowState.Maximized);
                    pane.MaximizePane(false);
                }
            }
        }

        public static void MakeSelectedTab(this DockableControlPane pane)
        {
            DockableGroupPane parent = pane.Parent;
            Debug.Assert(parent.ChildPaneStyle == ChildPaneStyle.TabGroup);
            if (parent.SelectedTabIndex != pane.Index)
                if (parent.Panes.Count > 1)
                    parent.SelectedTabIndex = pane.Index;
        }

        // Accounts for proxy form.
        public static bool IsPaneVisible(this DockableControlPane pane)
        {
            if (pane.IsPaneNormal())
            {
                return pane.IsVisible;
            }

            PaneProxyForm form = pane.GetProxyForm();
            return form.Visible;
        }

        // Gives focus to pane, accounting for maximized proxy form.
        public static void GiveFocus(this DockableControlPane pane)
        {
            Debug.Write("GiveFocus '" + pane.Text + "' ");
            if (pane.IsPaneNormal()) 
            {
                Debug.WriteLine("normal");
                pane.Control.Focus(); 
                return; 
            }

            PaneProxyForm form = pane.GetProxyForm();
            if (form.WindowState == FormWindowState.Maximized) 
            {
                Debug.WriteLine("maximized");
                form.Focus(); 
                return; 
            }

            if (form.Visible)
            {
                Debug.WriteLine("minimized and visible");
                form.Focus();
                return;
            }

            Debug.WriteLine("ignored");
        }

        private static void RepositionFloatingPane(this DockableControlPane pane, Point location, Size floatingSize)
        {
            // Adjust the size to account for chrome.
            Size captionSize = pane.Manager.ImageSizeCaption;
            int dx = SystemInformation.HorizontalResizeBorderThickness * 2;
            int dy = SystemInformation.VerticalResizeBorderThickness * 2 + captionSize.Height;
            Size adjustedSize = new Size(floatingSize.Width - dx, floatingSize.Height - dy);

            pane.DockAreaPane.FloatingLocation = location;
            pane.DockAreaPane.Size = adjustedSize;
        }

        public static void RepositionAsFloating(this DockableControlPane pane, Point location, Size floatingSize)
        {
            if (pane.DockedState == DockedState.Docked)
                pane.Float();

            RepositionFloatingPane(pane, location, floatingSize);
        }

        // Ensure the pane is floating and at z-order 0, i.e. in front of all other dock areas.
        // Disable events while resetting z-order to avoid responding to new active pane.
        public static void RepositionAsFloatingOnTop(this DockableControlPane pane, Point location, Size floatingSize)
        {
            if (pane.DockedState == DockedState.Docked) pane.Float();

            pane.WithoutEvents(() =>
                {
                    pane.Parent.Panes.Remove(pane);

                    // Create a new floating dock area pane and give it this control pane.
                    DockAreaPane dap = new DockAreaPane(DockedLocation.Floating);
                    dap.ChildPaneStyle = ChildPaneStyle.TabGroup;
                    dap.Panes.Add(pane);
                    //dap.FloatingLocation = new Point(int.MaxValue, int.MaxValue); // Offscreen
                    pane.Manager.DockAreas.Insert(dap, 0);
                });

            RepositionFloatingPane(pane, location, floatingSize);
        }

        public static void DestroyProxyPane(this DockableControlPane pane)
        {
            if (pane.HasProxyForm())
            {
                PaneProxyForm form = pane.GetProxyForm();
                form.Destroy();
                pane.UpdatePaneState((PaneProxyForm)null);
            }
        }

        public static void Destroy(this DockableControlPane pane)
        {
            pane.DestroyProxyPane();
            pane.Parent.Panes.Remove(pane);

            // According to Infragistics, the proper way to permanently dispose of a control pane
            // is to either dispose of the embedded control, or to remove the pane from 
            // DockManager.ControlPanes.
            pane.Control.Dispose();
        }

        public static void MoveToNewParent(this DockableControlPane pane, DockableGroupPane parent)
        {
            pane.Parent.Panes.Remove(pane);
            parent.Panes.Add(pane);
        }

        private static void WithoutEvents(this DockableControlPane pane, Action action)
        {
            DockEventManager dem = pane.Manager.EventManager;
            bool beforeEvents = dem.IsEnabled(DockManagerEventGroups.BeforeEvents);
            bool afterEvents = dem.IsEnabled(DockManagerEventGroups.AfterEvents);
            try
            {
                dem.SetEnabled(DockManagerEventGroups.AllEvents, false);
                action();
            }
            finally
            {
                dem.SetEnabled(DockManagerEventGroups.BeforeEvents, beforeEvents);
                dem.SetEnabled(DockManagerEventGroups.AfterEvents, afterEvents);
            }
        }
    }
}
