﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Infragistics.Win.UltraWinDock;

namespace WindowMaxMinDockMgr
{
    public partial class DockerPanel : UserControl
    {
        #region Member variables
        private readonly string DefaultDockAreaPaneKey = "DefaultDockAreaPane";

        private bool m_ShowAllDockIndicators = true;
        private bool m_ShowAllowedDockIndicatorsByValidation = false;
        private bool m_ShowAllowedDockIndicatorsByConfiguration = false;
        private bool m_HideWhenMinimized;
        #endregion Member variables

        #region Events and helper methods
        public event EventHandler DockChange;
        private void RaiseDockChangeEvent()
        {
            if (DockChange != null)
            {
                try
                {
                    DockChange(this, EventArgs.Empty);
                }
                catch (Exception ex)
                {
                    string msg = "Exception in RaiseDockChangeEvent: " + ex.GetType().FullName + ": " + ex.Message;
                    MessageBox.Show(msg);
                }
            }
        }

        public class PaneEventArgs : EventArgs
        {
            private readonly object _paneObj;
            public PaneEventArgs(object paneObj) { _paneObj = paneObj; }
            public object Pane { get { return _paneObj; } }
        }
        public delegate void PaneEventHandler(object sender, PaneEventArgs ea);
        public event PaneEventHandler PaneActivate;
        private void RaisePaneActivateEvent(object paneObj)
        {
            if (PaneActivate != null)
            {
                try
                {
                    PaneEventArgs pea = new PaneEventArgs(paneObj);
                    PaneActivate(this, pea);
                }
                catch (Exception ex)
                {
                    string msg = "Exception in RaisePaneActivateEvent: " + ex.GetType().FullName + ": " + ex.Message;
                    MessageBox.Show(msg);
                }
            }
        }

        #endregion Events

        #region Constructor
        public DockerPanel()
        {
            InitializeComponent();

            DefaultDockAreaLocation = DockedLocation.DockedBottom;

            // These are in designer-generated code, but I want them here to make intentions explicit,
            // i.e. so properties with default values are shown as well as those with non-default values.
            this.ultraDockManager1.CompressUnpinnedTabs = false;
            this.ultraDockManager1.DragWindowStyle = Infragistics.Win.UltraWinDock.DragWindowStyle.LayeredWindow;
            this.ultraDockManager1.HostControl = this;
            this.ultraDockManager1.LayoutStyle = Infragistics.Win.UltraWinDock.DockAreaLayoutStyle.FillContainer;
            this.ultraDockManager1.ShowMaximizeButton = false;
            this.ultraDockManager1.ShowMinimizeButton = false;
            this.ultraDockManager1.WindowStyle = Infragistics.Win.UltraWinDock.WindowStyle.VisualStudio2005;

            ShowAllDockIndicators();

            DockManager.AfterDockChange += new Infragistics.Win.UltraWinDock.PaneEventHandler(DockManager_AfterDockChange);
            DockManager.AfterToggleDockState += new Infragistics.Win.UltraWinDock.PaneEventHandler(DockManager_AfterToggleDockState);
            DockManager.PaneActivate += new ControlPaneEventHandler(DockManager_PaneActivate);
        }
        #endregion Constructor

        #region Properties
        public UltraDockManager DockManager { get { return ultraDockManager1; } }
        public DockedLocation DefaultDockAreaLocation { get; set; }
        public bool IsShowAllDockIndicators { get { return m_ShowAllDockIndicators; } }
        public bool IsShowAllowedDockIndicatorsByValidation { get { return m_ShowAllowedDockIndicatorsByValidation; } }
        public bool IsShowAllowedDockIndicatorsByConfiguration { get { return m_ShowAllowedDockIndicatorsByConfiguration; } }

        public bool HideMinimizedWindows
        {
            get { return m_HideWhenMinimized; }

            set
            {
                if (m_HideWhenMinimized != value)
                {
                    m_HideWhenMinimized = value;
                    WithEachControlPane((pane) =>
                    {
                        if (pane.HasProxyForm())
                        {
                            PaneProxyForm form = pane.GetProxyForm();
                            form.HideWhenMinimized = value;
                        }
                    });
                }
            }
        }
        private bool ShowWhenMinimized { get { return !HideMinimizedWindows; } }

        #endregion Properties

        #region Methods
        public XPanel CreatePanel()
        {
            XPanel panel = new XPanel();
            return panel;
        }

        public void DockPanel(UltraDockManager dm, XPanel panel, string caption)
        {
            CreateDefaultDockArea(dm);
            DockAreaPane da = GetDefaultDockArea(dm);
            Debug.Assert(da != null);

            DockableControlPane pane = new DockableControlPane(panel);
            pane.Text = caption;
            pane.TextTab = caption;
            pane.SetPaneState(PaneState.PaneStateValue.Normal, this);
            da.Panes.Add(pane);

            RaiseDockChangeEvent();
        }

        public bool IsDefaultDockArea(DockAreaPane dap) { return dap.Key != null && dap.Key == DefaultDockAreaPaneKey; }

        private void AllowDockAtLocation(DockablePaneBase pane, DockedLocation location)
        {
            switch (location)
            {
                case DockedLocation.DockedBottom:
                    pane.Settings.AllowDockBottom = Infragistics.Win.DefaultableBoolean.True;
                    break;

                case DockedLocation.DockedLeft:
                    pane.Settings.AllowDockLeft = Infragistics.Win.DefaultableBoolean.True;
                    break;

                case DockedLocation.DockedRight:
                    pane.Settings.AllowDockRight = Infragistics.Win.DefaultableBoolean.True;
                    break;

                case DockedLocation.DockedTop:
                    pane.Settings.AllowDockTop = Infragistics.Win.DefaultableBoolean.True;
                    break;
            }
        }

        private DockAreaPane GetDefaultDockArea(UltraDockManager dm)
        {
            return dm.PaneFromKey(DefaultDockAreaPaneKey) as DockAreaPane;
        }

        private void CreateDefaultDockArea(UltraDockManager dm)
        {
            Utilities.NestedTrace(() =>
            {
                DockAreaPane da = GetDefaultDockArea(dm);
                if (da == null)
                {
                    DockablePaneBase dpb = dm.PaneFromKey(DefaultDockAreaPaneKey);
                    if (dpb != null)
                    {
                        Debug.WriteLine("Pane exists with default dock area key, but not DockAreaPane. Strip key.");
                        dpb.Key = null;
                        dpb = dm.PaneFromKey(DefaultDockAreaPaneKey);
                    }
                    Debug.Assert(dpb == null);

                    Debug.WriteLine("No default dock area. Creating.");

                    da = new DockAreaPane(DefaultDockAreaLocation, DefaultDockAreaPaneKey);
                    if (da == null)
                    {
                        throw new ApplicationException("Cannot create default dock area");
                    }
                    da.ChildPaneStyle = ChildPaneStyle.TabGroup;
                    da.Key = DefaultDockAreaPaneKey;
                    dm.DockAreas.Add(da);
                }
            });
        }

        private void ReclaimDefaultDockArea(UltraDockManager dm)
        {
            Utilities.NestedTrace(() =>
            {
                Debug.Assert(GetDefaultDockArea(dm) == null);

                // If there is a dock area docked at the default location, make it the default dock area.
                foreach (DockAreaPane dap in dm.DockAreas)
                {
                    if (dap.DockedLocation == DefaultDockAreaLocation && dap.ChildPaneStyle == ChildPaneStyle.TabGroup)
                    {
                        dap.Key = DefaultDockAreaPaneKey;
                        Debug.WriteLine("Reclaimed dock area as default");
                        return;
                    }
                }
            });
        }

        private void FixupDefaultDockArea()
        {
            // If the default dock area isn't exactly how we want it, strip its status.
            DockAreaPane da = GetDefaultDockArea(DockManager);
            if (da != null)
            {
                if (da.DockedState == DockedState.Floating)
                {
                    Debug.WriteLine("Default dock area is floating! Setting Key = null.");
                    da.Key = null;
                }
                if (da.DockedLocation != DefaultDockAreaLocation)
                {
                    Debug.WriteLine("Default dock area not in default location! Setting Key = null.");
                    da.Key = null;
                }
                da = GetDefaultDockArea(DockManager);
            }

            // If the default dock area was changed to a group pane hosted within another,
            // strip its status.
            if (da == null)
            {
                DockableGroupPane dgp = DockManager.PaneFromKey(DefaultDockAreaPaneKey) as DockableGroupPane;
                if (dgp != null)
                {
                    Debug.WriteLine("Found default dock area as internal node. Setting Key = null.");
                    dgp.Key = null;
                }

                Debug.WriteLine("No default dock area. Try to reclaim.");
                ReclaimDefaultDockArea(DockManager);
                da = GetDefaultDockArea(DockManager);
            }

            if (da == null)
            {
                Debug.WriteLine("No default dock area. Try to create.");
                CreateDefaultDockArea(DockManager);
                da = GetDefaultDockArea(DockManager);
            }
            Debug.Assert(da != null);
        }

        // Eliminate unwanted dock areas. We do this indirectly by moving panes to
        // the desired places, and letting the dock manager garbage-collect when it 
        // pleases (usually on the next caption or tab double-click).
        //
        // for each control pane
        //      if part of default dock area, leave alone
        //      if part of a floating tab group, leave alone
        //      remove from parent group and add to default dock area
        private void FixupControlPane(DockableControlPane pane)
        {
            DockAreaPane da = GetDefaultDockArea(DockManager);
            Debug.Assert(da != null);

            if (pane.DockAreaPane == da)
            {
                Debug.WriteLine("Control pane '" + pane.Text + "' is docked in default area -- OK");
                return;
            }

            if (pane.DockedState == DockedState.Floating)
            {
                Debug.WriteLine("Control pane '" + pane.Text + "' is floating");

                if (pane.Parent == pane.DockAreaPane)
                {
                    Debug.WriteLine("Parent is floating dock area -- OK");
                    return;
                }

                Debug.WriteLine("Parent is not floating dock area -- restructuring");

                DockAreaPane dap = pane.DockAreaPane;
                dap.ChildPaneStyle = ChildPaneStyle.TabGroup;
                pane.MoveToNewParent(dap);
                return;
            }

            Debug.WriteLine("Control pane '" + pane.Text + "' is docked in non-default area -- moving to default area");
            pane.Parent.Panes.Remove(pane);
            da.Panes.Add(pane);
        }
        
        private void FixupDockStructure()
        {
            Utilities.NestedTrace(() =>
            {
                WithoutEvents(() =>
                {
                    FixupDefaultDockArea();
                    WithEachControlPane(FixupControlPane);
                });
            });
        }

        private void WithoutEvents(Action action)
        {
            DockEventManager dem = DockManager.EventManager;
            bool beforeEvents = dem.IsEnabled(DockManagerEventGroups.BeforeEvents);
            bool afterEvents = dem.IsEnabled(DockManagerEventGroups.AfterEvents);
            try
            {
                dem.SetEnabled(DockManagerEventGroups.AllEvents, false);
                action();
            }
            finally
            {
                dem.SetEnabled(DockManagerEventGroups.BeforeEvents, beforeEvents);
                dem.SetEnabled(DockManagerEventGroups.AfterEvents, afterEvents);
            }
        }

        public void WithEachTopLevelPane(Action<object> action)
        {
            foreach (DockablePaneBase pane in DockManager.DockAreas) action(pane);
        }

        public void WithEachPane(Action<DockablePaneBase> action)
        {
            foreach (DockablePaneBase pane in DockManager.DockAreas) WithEachSubPane(pane, action);
        }

        // Iterates over children of initial pane, applying action to each.
        public void WithEachChildPane(object paneObj, Action<object> action)
        {
            Debug.WriteLine("WithEachSubPane '" + GetPaneText(paneObj) + "'");
            DockableGroupPane group = paneObj as DockableGroupPane;
            if (group == null) return;
            foreach (DockablePaneBase childPane in group.Panes) action(childPane);
        }

        // Applies action to root pane, then recurses through structure appling action to each subpane.
        private void WithEachSubPane(DockablePaneBase pane, Action<DockablePaneBase> action)
        {
            Debug.WriteLine("WithEachSubPane '" + pane.Text + "'");
            action(pane);
            DockableGroupPane group = pane as DockableGroupPane;
            if (group == null) return;
            foreach (DockablePaneBase childPane in group.Panes) WithEachSubPane(childPane, action);
        }

        public void WithEachControlPane(Action<DockableControlPane> action)
        {
            foreach (DockableControlPane pane in DockManager.ControlPanes) action(pane);
        }

        public void ShowAllDockIndicators()
        {
            Debug.WriteLine("ShowAllDockIndicators");

            m_ShowAllDockIndicators = true;
            m_ShowAllowedDockIndicatorsByValidation = false;
            m_ShowAllowedDockIndicatorsByConfiguration = false;

            DockManager.DefaultPaneSettings.AllowDockAsTab = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowDockBottom = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowDockLeft = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowDockTop = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowDockRight = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowDragging = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowFloating = Infragistics.Win.DefaultableBoolean.True;

            WithEachPane((pane) =>
            {
                pane.ResetSettings();
                DumpResolvedDockSettings(pane);
            });

            DockManager.BeforeDockChange -= new BeforeDockChangeEventHandler(DockManager_BeforeDockChange);
        }

        public void ShowAllowedDockIndicatorsByValidation()
        {
            Debug.WriteLine("ShowAllowedDockIndicatorsByValidation");

            m_ShowAllDockIndicators = false;
            m_ShowAllowedDockIndicatorsByValidation = true;
            m_ShowAllowedDockIndicatorsByConfiguration = false;

            DockManager.DefaultPaneSettings.AllowDockAsTab = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowDockBottom = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowDockLeft = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowDockTop = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowDockRight = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowDragging = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowFloating = Infragistics.Win.DefaultableBoolean.True;

            WithEachPane((pane) =>
            {
                pane.ResetSettings();
                DumpResolvedDockSettings(pane);
            });

            DockManager.BeforeDockChange += new BeforeDockChangeEventHandler(DockManager_BeforeDockChange);
        }

        public void ShowAllowedDockIndicatorsByConfiguration()
        {
            Debug.WriteLine("ShowAllowedDockIndicatorsByConfiguration");

            m_ShowAllDockIndicators = false;
            m_ShowAllowedDockIndicatorsByValidation = false;
            m_ShowAllowedDockIndicatorsByConfiguration = true;

            DockManager.DefaultPaneSettings.AllowDockAsTab = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowDockBottom = Infragistics.Win.DefaultableBoolean.False;
            DockManager.DefaultPaneSettings.AllowDockLeft = Infragistics.Win.DefaultableBoolean.False;
            DockManager.DefaultPaneSettings.AllowDockTop = Infragistics.Win.DefaultableBoolean.False;
            DockManager.DefaultPaneSettings.AllowDockRight = Infragistics.Win.DefaultableBoolean.False;
            DockManager.DefaultPaneSettings.AllowDragging = Infragistics.Win.DefaultableBoolean.True;
            DockManager.DefaultPaneSettings.AllowFloating = Infragistics.Win.DefaultableBoolean.True;

            WithEachPane((pane) =>
            {
                pane.ResetSettings();
                AllowDockAtLocation(pane, DefaultDockAreaLocation);
                DumpResolvedDockSettings(pane);
            });


            DockManager.BeforeDockChange -= new BeforeDockChangeEventHandler(DockManager_BeforeDockChange);
        }

        private void DumpResolvedDockSettings(DockablePaneBase pane)
        {
            Debug.Indent();
            Debug.WriteLine(FormatDockSetting("AllowDockBottom", pane.SettingsResolved.AllowDockBottom));
            Debug.WriteLine(FormatDockSetting("AllowDockLeft", pane.SettingsResolved.AllowDockLeft));
            Debug.WriteLine(FormatDockSetting("AllowDockRight", pane.SettingsResolved.AllowDockRight));
            Debug.WriteLine(FormatDockSetting("AllowDockTop", pane.SettingsResolved.AllowDockTop));
            Debug.Unindent();
        }

        private string FormatDockSetting(string name, Infragistics.Win.DefaultableBoolean value)
        {
            string result = name + " ";
            if (value == Infragistics.Win.DefaultableBoolean.Default) result += "default";
            else if (value == Infragistics.Win.DefaultableBoolean.True) result += "true";
            else result += "false";
            return result;
        }

        public void CloseAllPanes() { WithEachControlPane((pane) => { pane.DestroyProxyPane(); }); }

        public void TilePanes(List<object> panes, Point pt, Size sz,
            Func<int, Point, Size, List<Tuple<Point, Size>>> getPositions)
        {
            List<DockableControlPane> controlPanes = new List<DockableControlPane>();
            foreach (object paneObj in panes)
            {
                DockableControlPane controlPane = paneObj as DockableControlPane;
                Debug.Assert(controlPane != null);
                controlPanes.Add(controlPane);
            }
            TilePanes(controlPanes, pt, sz, getPositions);
        }

        public void TilePanes(List<DockableControlPane> panes, Point pt, Size sz,
            Func<int, Point, Size, List<Tuple<Point, Size>>> getPositions)
        {
            var positions = getPositions(panes.Count, pt, sz);
            var panesWithPositions = Utilities.Zip(panes, positions);
            RepositionFloatingPanes(panesWithPositions);
        }

        // Panes will be floating afterwards.
        private void RepositionFloatingPanes(
            List<Tuple<DockableControlPane, Tuple<Point, Size>>> panesWithPositions)
        {
            for (int i = panesWithPositions.Count() - 1; i >= 0; --i)
            {
                var paneWithPosition = panesWithPositions[i];
                DockableControlPane pane = paneWithPosition.Item1;
                Point pt = paneWithPosition.Item2.Item1;
                Size sz = paneWithPosition.Item2.Item2;
                pane.RepositionAsFloatingOnTop(pt, sz);
            }
        }

        private DockableControlPane CastToDCP(object paneObj)
        {
            DockableControlPane pane = paneObj as DockableControlPane;
            Debug.Assert(pane != null, Utilities.GetCallingMethodName(2) + " argument must be DockableControlPanel");
            return pane;
        }

        public void SelectTab(object paneObj)
        {
            Debug.WriteLine("DockerPanel.SelectTab");
            if (paneObj == null) return;

            DockableControlPane pane = CastToDCP(paneObj);
            Debug.WriteLine("\t" + pane.Text);
            pane.MakeSelectedTab();
        }

        public object GetActivePane() { return this.DockManager.ActivePane; }
        public bool IsPaneMinimized(object paneObj) { return CastToDCP(paneObj).IsPaneMinimized(); }
        public bool IsPaneMaximized(object paneObj) { return CastToDCP(paneObj).IsPaneMaximized(); }
        public bool IsPaneNormal(object paneObj) { return CastToDCP(paneObj).IsPaneNormal(); }
        public bool IsPaneVisible(object paneObj) { return CastToDCP(paneObj).IsPaneVisible(); }
        public bool IsPaneDocked(object paneObj) { return CastToDCP(paneObj).DockedState == DockedState.Docked; }
        public bool PaneHasProxyForm(object paneObj) { return CastToDCP(paneObj).HasProxyForm(); }
        public PaneProxyForm GetProxyForm(object paneObj) { return CastToDCP(paneObj).GetProxyForm(); }
        public void GiveFocus(object paneObj) { CastToDCP(paneObj).GiveFocus(); }
        public void ToggleDockState(object paneObj) { CastToDCP(paneObj).ToggleDockState(); }
        public void RestorePane(object paneObj) { CastToDCP(paneObj).RestorePane(); }
        public void MinimizePane(object paneObj, bool showForm) { CastToDCP(paneObj).MinimizePane(showForm); }
        public void MaximizePane(object paneObj, bool showForm) { CastToDCP(paneObj).MaximizePane(showForm); }
        public void DestroyPane(object paneObj) { CastToDCP(paneObj).Destroy(); }
        private string GetControlPaneText(DockableControlPane pane) { return pane.Text; }

        public bool IsEmptyGroup(object paneObj)
        {
            DockableGroupPane group = paneObj as DockableGroupPane;
            if (group == null) return false;
            foreach (DockablePaneBase childPane in group.Panes)
            {
                if (!IsEmptyGroup(childPane)) return false;
            }
            return true;
        }

        public string GetPaneText(object paneObj)
        {
            DockableGroupPane groupPane = paneObj as DockableGroupPane;
            if (groupPane == null)
            {
                // Leaf node
                DockableControlPane controlPane = paneObj as DockableControlPane;
                return GetControlPaneText(controlPane);
            }

            // Add text indicating how pane is docked relative to parent.
            return GetGroupPaneText(groupPane);
        }

        private string GetGroupPaneText(DockableGroupPane pane)
        {
            // "Default DockArea DockBottom Docked TabGroup"

            StringBuilder sb = new StringBuilder();

            if (!string.IsNullOrEmpty(pane.Key)) sb.Append(pane.Key);

            DockAreaPane dap = pane as DockAreaPane;
            if (dap != null)
            {
                if (sb.Length > 0) sb.Append(" ");
                sb.Append(Enum.GetName(typeof(DockedLocation), dap.DockedLocation));
            }

            if (sb.Length > 0) sb.Append(" ");
            sb.Append(Enum.GetName(typeof(ChildPaneStyle), pane.ChildPaneStyle));

            return sb.ToString();
        }
        #endregion Methods

        #region Event handlers
        private void DockManager_AfterToggleDockState(object sender, Infragistics.Win.UltraWinDock.PaneEventArgs e)
        {
            FixupDockStructure();
            RaiseDockChangeEvent();
        }

        private void DockManager_AfterDockChange(object sender, Infragistics.Win.UltraWinDock.PaneEventArgs e)
        {
            FixupDockStructure();
            RaiseDockChangeEvent();
        }

        static string last_msg = "";
        void DockManager_BeforeDockChange(object sender, BeforeDockChangeEventArgs e)
        {
            string msg = "DockManager_BeforeDockChange: ";
            msg += Enum.GetName(typeof(DockChangeType), e.ChangeType);
            msg += " ";
            msg += Enum.GetName(typeof(DockedLocation), e.NewDockedLocation);
            msg += " ";
            msg += "Index=" + e.NewIndex.ToString();
            msg += " ";

            switch (e.ChangeType)
            {
                case DockChangeType.FloatingWindowReposition:
                case DockChangeType.SiblingPaneReposition:
                case DockChangeType.NewSiblingPane:
                    msg += "OK";
                    break;

                case DockChangeType.NewDockArea:
                    if ((e.NewDockedLocation != DefaultDockAreaLocation) &&
                       (e.NewDockedLocation != DockedLocation.Floating))
                    {
                        msg += "Cancelled -- not default or floating";
                        e.Cancel = true;
                    }
                    else
                    {
                        msg += "OK";
                    }
                    break;

                case DockChangeType.DisplayAsMdiChildren:
                case DockChangeType.NewGroup:
                    e.Cancel = true;
                    msg += "Cancelled";
                    break;
            }
            if (last_msg != msg)
            {
                Debug.WriteLine(msg);
                last_msg = msg;
            }
        }

        void DockManager_PaneActivate(object sender, ControlPaneEventArgs e)
        {
            Utilities.NestedTrace(() =>
            {
                RaisePaneActivateEvent(e.Pane);
            });
        }
        #endregion Event handlers
    }
}
